# 15 "src/ocaml/preprocess/408/lexer_raw.mll"
 
open Std
open Misc
open Lexing
open Parser_raw

type keywords = (string, Parser_raw.token) Hashtbl.t

type error =
  | Illegal_character of char
  | Illegal_escape of string * string option
  | Reserved_sequence of string * string option
  | Unterminated_comment of Location.t
  | Unterminated_string
  | Unterminated_string_in_comment of Location.t * Location.t
  | Keyword_as_label of string
  | Invalid_literal of string

exception Error of error * Location.t

(* Monad in which the lexer evaluates *)
type 'a result =
  | Return of 'a
  | Refill of (unit -> 'a result)
  | Fail of error * Location.t

let return a = Return a

let fail lexbuf e = Fail (e, Location.curr lexbuf)
let fail_loc e l = Fail (e,l)

let rec (>>=) (m : 'a result) (f : 'a -> 'b result) : 'b result =
  match m with
  | Return a -> f a
  | Refill u ->
    Refill (fun () -> u () >>= f)
  | Fail _ as e -> e

type preprocessor = (Lexing.lexbuf -> Parser_raw.token) -> Lexing.lexbuf -> Parser_raw.token

type state = {
  keywords: keywords;
  mutable buffer: Buffer.t;
  mutable string_start_loc: Location.t;
  mutable comment_start_loc: Location.t list;
  mutable preprocessor: preprocessor option;
}

let make ?preprocessor keywords = {
  keywords;
  buffer = Buffer.create 17;
  string_start_loc = Location.none;
  comment_start_loc = [];
  preprocessor;
}

let lABEL m = m >>= fun v -> return (LABEL v)
let oPTLABEL m = m >>= fun v -> return (OPTLABEL v)

let rec catch m f = match m with
  | Fail (e,l) -> f e l
  | Refill next -> Refill (fun () -> catch (next ()) f)
  | Return _ -> m

(* The table of keywords *)

let keyword_table : keywords =
  create_hashtable 149 [
    "and", AND;
    "as", AS;
    "assert", ASSERT;
    "begin", BEGIN;
    "class", CLASS;
    "constraint", CONSTRAINT;
    "do", DO;
    "done", DONE;
    "downto", DOWNTO;
    "else", ELSE;
    "end", END;
    "exception", EXCEPTION;
    "external", EXTERNAL;
    "false", FALSE;
    "for", FOR;
    "fun", FUN;
    "function", FUNCTION;
    "functor", FUNCTOR;
    "if", IF;
    "in", IN;
    "include", INCLUDE;
    "inherit", INHERIT;
    "initializer", INITIALIZER;
    "lazy", LAZY;
    "let", LET;
    "match", MATCH;
    "method", METHOD;
    "module", MODULE;
    "mutable", MUTABLE;
    "new", NEW;
    "nonrec", NONREC;
    "object", OBJECT;
    "of", OF;
    "open", OPEN;
    "or", OR;
(*  "parser", PARSER; *)
    "private", PRIVATE;
    "rec", REC;
    "sig", SIG;
    "struct", STRUCT;
    "then", THEN;
    "to", TO;
    "true", TRUE;
    "try", TRY;
    "type", TYPE;
    "val", VAL;
    "virtual", VIRTUAL;
    "when", WHEN;
    "while", WHILE;
    "with", WITH;

    "lor", INFIXOP3("lor"); (* Should be INFIXOP2 *)
    "lxor", INFIXOP3("lxor"); (* Should be INFIXOP2 *)
    "mod", INFIXOP3("mod");
    "land", INFIXOP3("land");
    "lsl", INFIXOP4("lsl");
    "lsr", INFIXOP4("lsr");
    "asr", INFIXOP4("asr");
]

let keywords l = create_hashtable 11 l

(* To store the position of the beginning of a string and comment *)
let in_comment state = state.comment_start_loc <> []

(* Escaped chars are interpreted in strings unless they are in comments. *)
let store_escaped_uchar state lexbuf u =
  if in_comment state
  then Buffer.add_string state.buffer (Lexing.lexeme lexbuf)
  else Buffer.add_utf_8_uchar state.buffer u

(* to translate escape sequences *)

let digit_value c =
  match c with
  | 'a' .. 'f' -> 10 + Char.code c - Char.code 'a'
  | 'A' .. 'F' -> 10 + Char.code c - Char.code 'A'
  | '0' .. '9' -> Char.code c - Char.code '0'
  | _ -> assert false

let num_value lexbuf ~base ~first ~last =
  let c = ref 0 in
  for i = first to last do
    let v = digit_value (Lexing.lexeme_char lexbuf i) in
    assert(v < base);
    c := (base * !c) + v
  done;
  !c

let char_for_backslash = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c   -> c

let illegal_escape lexbuf reason =
  let error = Illegal_escape (Lexing.lexeme lexbuf, Some reason) in
  fail lexbuf error

let char_for_decimal_code state lexbuf i =
  let c = num_value lexbuf ~base:10 ~first:i ~last:(i+2) in
  if (c < 0 || c > 255) then
    if in_comment state
    then return 'x'
    else
      illegal_escape lexbuf
        (Printf.sprintf
          "%d is outside the range of legal characters (0-255)." c)
  else return (Char.chr c)

let char_for_octal_code state lexbuf i =
  let c = num_value lexbuf ~base:8 ~first:i ~last:(i+2) in
  if (c < 0 || c > 255) then
    if in_comment state
    then return 'x'
    else
      illegal_escape lexbuf
        (Printf.sprintf
          "o%o (=%d) is outside the range of legal characters (0-255)." c c)
  else return (Char.chr c)

let char_for_hexadecimal_code lexbuf i =
  Char.chr (num_value lexbuf ~base:16 ~first:i ~last:(i+1))

let uchar_for_uchar_escape lexbuf =
  let illegal_escape lexbuf reason =
    let error = Illegal_escape (Lexing.lexeme lexbuf, Some reason) in
    raise (Error (error, Location.curr lexbuf))
  in
  let len = Lexing.lexeme_end lexbuf - Lexing.lexeme_start lexbuf in
  let first = 3 (* skip opening \u{ *) in
  let last = len - 2 (* skip closing } *) in
  let digit_count = last - first + 1 in
  match digit_count > 6 with
  | true ->
      illegal_escape lexbuf
        "too many digits, expected 1 to 6 hexadecimal digits"
  | false ->
      let cp = num_value lexbuf ~base:16 ~first ~last in
      if Uchar.is_valid cp then Uchar.unsafe_of_int cp else
      illegal_escape lexbuf
        (Printf.sprintf "%X is not a Unicode scalar value" cp)

let keyword_or state s default =
  try Hashtbl.find state.keywords s
      with Not_found -> try Hashtbl.find keyword_table s
  with Not_found -> default

let is_keyword name = Hashtbl.mem keyword_table name

let check_label_name lexbuf name =
  if is_keyword name
  then fail lexbuf (Keyword_as_label name)
  else return name

(* Update the current location with file name and line number. *)

let update_loc lexbuf _file line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_file = pos.pos_fname
    (*match file with
      | None -> pos.pos_fname
      | Some s -> s*)
  in
  lexbuf.lex_curr_p <- { pos with
    pos_fname = new_file;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(* Warn about Latin-1 characters used in idents *)

let warn_latin1 lexbuf =
  Location.deprecated (Location.curr lexbuf)
    "ISO-Latin1 characters in identifiers"
;;

(* Error report *)

open Format

let prepare_error loc = function
  | Illegal_character c ->
      Location.errorf ~loc "Illegal character (%s)" (Char.escaped c)
  | Illegal_escape (s, explanation) ->
      Location.errorf ~loc
        "Illegal backslash escape in string or character (%s)%t" s
        (fun ppf -> match explanation with
           | None -> ()
           | Some expl -> fprintf ppf ": %s" expl)
  | Reserved_sequence (s, explanation) ->
      Location.errorf ~loc
        "Reserved character sequence: %s%t" s
        (fun ppf -> match explanation with
           | None -> ()
           | Some expl -> fprintf ppf " %s" expl)
  | Unterminated_comment _ ->
      Location.errorf ~loc "Comment not terminated"
  | Unterminated_string ->
      Location.errorf ~loc "String literal not terminated"
  | Unterminated_string_in_comment (_, literal_loc) ->
      Location.errorf ~loc
        "This comment contains an unterminated string literal"
        ~sub:[Location.msg ~loc:literal_loc "String literal begins here"]
  | Keyword_as_label kwd ->
      Location.errorf ~loc
        "`%s' is a keyword, it cannot be used as label name" kwd
  | Invalid_literal s ->
      Location.errorf ~loc "Invalid literal %s" s
(* FIXME: Invalid_directive? *)

let () =
  Location.register_error_of_exn
    (function
      | Error (err, loc) ->
        Some (prepare_error loc err)
      | _ ->
        None
    )


# 294 "src/ocaml/preprocess/lexer_raw.ml"
let __ocaml_lex_refill : (Lexing.lexbuf -> 'a) -> (Lexing.lexbuf -> 'a) =
# 349 "src/ocaml/preprocess/408/lexer_raw.mll"
       (fun k lexbuf -> Refill (fun () -> k lexbuf))
# 298 "src/ocaml/preprocess/lexer_raw.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\158\255\159\255\224\000\003\001\038\001\073\001\108\001\
    \143\001\185\255\178\001\193\255\091\000\215\001\250\001\068\000\
    \071\000\029\002\211\255\213\255\216\255\064\002\159\002\194\002\
    \088\000\255\000\224\002\232\255\093\003\221\003\099\004\227\004\
    \105\005\233\005\111\006\239\006\025\007\248\007\248\008\117\007\
    \084\008\122\000\254\255\001\000\005\000\255\255\006\000\007\000\
    \207\255\029\009\250\255\252\255\184\255\186\255\064\009\099\009\
    \249\009\121\010\247\255\246\255\255\010\127\011\164\011\243\255\
    \242\255\037\012\165\012\109\002\043\013\171\013\208\013\081\014\
    \209\014\107\015\235\015\113\016\241\016\119\017\247\017\208\018\
    \080\019\234\019\087\000\106\020\240\020\112\021\010\022\138\022\
    \108\000\191\255\231\255\192\002\235\022\106\000\107\000\011\000\
    \230\255\229\255\224\255\099\004\118\000\099\000\113\000\228\255\
    \128\000\136\000\208\000\227\255\224\000\219\000\226\255\227\004\
    \220\000\225\255\222\000\222\255\235\022\221\255\018\023\161\005\
    \127\002\220\255\012\000\024\001\034\001\045\001\043\001\220\255\
    \013\000\064\023\103\023\138\023\176\023\203\255\204\255\205\255\
    \201\255\211\023\227\000\219\000\194\255\195\255\196\255\225\000\
    \181\255\179\255\188\255\246\023\025\024\060\024\095\024\130\024\
    \175\004\243\255\244\255\017\000\245\255\047\001\055\008\253\255\
    \250\000\000\001\255\255\254\255\252\255\085\008\117\009\004\001\
    \012\001\018\000\251\255\250\255\249\255\233\005\194\002\013\001\
    \248\255\211\002\020\001\247\255\128\006\030\001\246\255\060\001\
    \104\001\246\255\247\255\248\255\077\001\186\024\255\255\249\255\
    \205\000\117\007\221\002\253\255\073\001\094\001\111\001\122\004\
    \252\255\249\009\251\255\121\010\226\014\250\255\112\001\130\001\
    \252\255\219\007\254\255\255\255\113\001\139\001\253\255\162\011\
    \043\001\055\001\079\001\098\001\060\001\103\001\060\001\019\000\
    \255\255";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\091\000\090\000\087\000\086\000\079\000\
    \077\000\255\255\065\000\255\255\058\000\057\000\055\000\053\000\
    \049\000\082\000\255\255\255\255\255\255\037\000\036\000\043\000\
    \041\000\040\000\063\000\255\255\018\000\018\000\017\000\016\000\
    \015\000\014\000\014\000\014\000\010\000\007\000\006\000\068\000\
    \046\000\002\000\255\255\097\000\097\000\255\255\255\255\255\255\
    \255\255\047\000\255\255\255\255\255\255\255\255\004\000\085\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\011\000\255\255\
    \255\255\014\000\014\000\094\000\014\000\014\000\095\000\022\000\
    \022\000\020\000\019\000\022\000\019\000\019\000\018\000\020\000\
    \019\000\020\000\255\255\021\000\021\000\018\000\018\000\020\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\031\000\031\000\031\000\031\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\032\000\255\255\089\000\255\255\092\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\038\000\088\000\083\000\045\000\255\255\255\255\255\255\
    \255\255\056\000\075\000\072\000\255\255\255\255\255\255\073\000\
    \255\255\255\255\255\255\066\000\084\000\078\000\081\000\080\000\
    \255\255\255\255\255\255\012\000\255\255\012\000\012\000\255\255\
    \012\000\012\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\009\000\009\000\255\255\255\255\
    \006\000\006\000\006\000\255\255\001\000\006\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\003\000\255\255\255\255\255\255\
    \002\000\255\255\255\255\001\000\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\093\000\255\255\000\000\071\000\071\000\030\000\031\000\
    \032\000\033\000\033\000\033\000\255\255\255\255\033\000\255\255\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\000\000\255\255\255\255\
    \056\000\057\000\000\000\000\000\060\000\061\000\255\255\000\000\
    \000\000\033\000\033\000\255\255\033\000\033\000\255\255\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\255\255\071\000\071\000\071\000\071\000\071\000\
    \255\255\000\000\000\000\255\255\098\000\255\255\255\255\255\255\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\000\000\255\255\000\000\255\255\123\000\
    \255\255\000\000\255\255\123\000\124\000\123\000\126\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \153\000\000\000\000\000\255\255\000\000\167\000\255\255\000\000\
    \255\255\255\255\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\255\255\255\255\255\255\
    \000\000\255\255\255\255\000\000\255\255\255\255\000\000\255\255\
    \185\000\000\000\000\000\000\000\255\255\191\000\000\000\000\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\255\255\255\255\000\000\255\255\208\000\
    \000\000\255\255\000\000\000\000\255\255\255\255\000\000\255\255\
    \255\255\255\255\218\000\221\000\255\255\221\000\255\255\255\255\
    \000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\041\000\042\000\042\000\041\000\043\000\047\000\045\000\
    \045\000\042\000\046\000\046\000\047\000\094\000\121\000\127\000\
    \095\000\122\000\128\000\154\000\168\000\224\000\183\000\169\000\
    \041\000\008\000\027\000\022\000\006\000\004\000\021\000\025\000\
    \024\000\019\000\023\000\007\000\018\000\017\000\040\000\003\000\
    \029\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\016\000\015\000\014\000\013\000\039\000\036\000\
    \005\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\012\000\044\000\011\000\005\000\038\000\
    \020\000\034\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\035\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\026\000\010\000\009\000\037\000\136\000\
    \138\000\135\000\114\000\041\000\134\000\133\000\041\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\097\000\096\000\104\000\104\000\104\000\104\000\141\000\
    \103\000\140\000\041\000\139\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\108\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\
    \105\000\105\000\105\000\105\000\105\000\105\000\105\000\105\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\142\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\107\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \002\000\003\000\110\000\113\000\003\000\003\000\003\000\115\000\
    \145\000\094\000\003\000\003\000\095\000\003\000\003\000\003\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\003\000\143\000\003\000\003\000\003\000\003\000\
    \003\000\144\000\121\000\163\000\003\000\122\000\255\255\003\000\
    \003\000\003\000\162\000\172\000\121\000\003\000\003\000\122\000\
    \003\000\003\000\003\000\171\000\176\000\127\000\125\000\121\000\
    \128\000\168\000\122\000\179\000\169\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\126\000\182\000\154\000\005\000\
    \203\000\183\000\005\000\005\000\005\000\125\000\217\000\124\000\
    \005\000\005\000\196\000\005\000\005\000\005\000\170\000\187\000\
    \218\000\219\000\206\000\092\000\003\000\222\000\003\000\223\000\
    \005\000\003\000\005\000\005\000\005\000\005\000\005\000\000\000\
    \196\000\196\000\006\000\198\000\220\000\006\000\006\000\006\000\
    \000\000\220\000\187\000\006\000\006\000\188\000\006\000\006\000\
    \006\000\196\000\187\000\211\000\198\000\206\000\213\000\003\000\
    \000\000\003\000\000\000\006\000\005\000\006\000\006\000\006\000\
    \006\000\006\000\190\000\166\000\211\000\130\000\000\000\212\000\
    \130\000\130\000\130\000\000\000\000\000\211\000\130\000\130\000\
    \213\000\130\000\151\000\130\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\000\000\130\000\006\000\
    \130\000\150\000\130\000\130\000\130\000\000\000\000\000\000\000\
    \148\000\000\000\000\000\148\000\148\000\148\000\000\000\000\000\
    \000\000\148\000\148\000\000\000\148\000\148\000\148\000\000\000\
    \000\000\000\000\000\000\000\000\189\000\006\000\000\000\006\000\
    \000\000\148\000\130\000\148\000\149\000\148\000\148\000\148\000\
    \000\000\000\000\000\000\006\000\000\000\000\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\006\000\006\000\000\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \130\000\000\000\130\000\000\000\006\000\148\000\006\000\006\000\
    \006\000\006\000\006\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\000\000\000\000\006\000\006\000\006\000\209\000\255\255\
    \000\000\006\000\006\000\000\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\000\000\148\000\000\000\148\000\000\000\146\000\
    \006\000\006\000\000\000\006\000\006\000\006\000\006\000\006\000\
    \255\255\000\000\000\000\006\000\000\000\000\000\006\000\006\000\
    \006\000\000\000\255\255\000\000\006\000\006\000\000\000\137\000\
    \006\000\006\000\000\000\255\255\000\000\255\255\147\000\255\255\
    \006\000\000\000\000\000\000\000\006\000\006\000\006\000\006\000\
    \006\000\006\000\006\000\000\000\000\000\000\000\130\000\000\000\
    \000\000\130\000\130\000\130\000\000\000\000\000\000\000\130\000\
    \130\000\000\000\130\000\131\000\130\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\006\000\000\000\006\000\000\000\130\000\
    \006\000\130\000\130\000\132\000\130\000\130\000\000\000\000\000\
    \000\000\006\000\255\255\000\000\006\000\006\000\129\000\255\255\
    \186\000\000\000\006\000\006\000\000\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\006\000\000\000\
    \006\000\000\000\006\000\130\000\006\000\006\000\006\000\006\000\
    \006\000\000\000\210\000\000\000\000\000\000\000\000\000\000\000\
    \120\000\000\000\000\000\000\000\000\000\000\000\067\000\000\000\
    \000\000\067\000\067\000\067\000\000\000\000\000\000\000\067\000\
    \067\000\130\000\067\000\130\000\067\000\000\000\006\000\120\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\067\000\
    \120\000\000\000\067\000\067\000\067\000\067\000\000\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\000\000\000\000\000\000\006\000\000\000\006\000\120\000\
    \118\000\000\000\118\000\118\000\118\000\118\000\000\000\000\000\
    \000\000\118\000\118\000\067\000\118\000\118\000\118\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\118\000\000\000\118\000\118\000\118\000\118\000\118\000\
    \000\000\000\000\000\000\003\000\000\000\000\000\003\000\003\000\
    \003\000\067\000\000\000\117\000\116\000\003\000\000\000\003\000\
    \003\000\003\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\003\000\118\000\003\000\003\000\
    \003\000\003\000\003\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\199\000\199\000\199\000\
    \199\000\199\000\199\000\199\000\199\000\199\000\199\000\000\000\
    \000\000\000\000\000\000\118\000\089\000\118\000\000\000\091\000\
    \003\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\000\000\090\000\000\000\003\000\091\000\
    \003\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\000\000\090\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\073\000\255\255\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\072\000\000\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \255\255\255\255\255\255\255\255\028\000\255\255\000\000\000\000\
    \000\000\000\000\072\000\000\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\073\000\255\255\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\075\000\
    \000\000\000\000\072\000\000\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\076\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\077\000\074\000\074\000\
    \255\255\255\255\255\255\255\255\028\000\255\255\000\000\075\000\
    \000\000\000\000\072\000\000\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\076\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\077\000\074\000\074\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\111\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\111\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\111\000\111\000\111\000\111\000\
    \111\000\111\000\200\000\200\000\200\000\200\000\200\000\200\000\
    \200\000\200\000\200\000\200\000\000\000\000\000\000\000\000\000\
    \000\000\154\000\000\000\000\000\155\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\111\000\111\000\111\000\111\000\
    \111\000\111\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\159\000\000\000\000\000\000\000\000\000\157\000\161\000\
    \000\000\160\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\112\000\112\000\112\000\112\000\112\000\
    \112\000\112\000\112\000\112\000\112\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\112\000\112\000\112\000\112\000\
    \112\000\112\000\158\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\112\000\112\000\112\000\112\000\
    \112\000\112\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\125\000\121\000\000\000\000\000\122\000\156\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\125\000\000\000\124\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\180\000\180\000\180\000\180\000\180\000\180\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\180\000\180\000\180\000\180\000\180\000\180\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \181\000\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\181\000\181\000\181\000\181\000\181\000\181\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\068\000\000\000\000\000\
    \000\000\181\000\181\000\181\000\181\000\181\000\181\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\055\000\000\000\000\000\055\000\055\000\055\000\
    \000\000\000\000\000\000\055\000\055\000\000\000\055\000\055\000\
    \055\000\000\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\000\000\000\000\055\000\065\000\055\000\055\000\055\000\
    \062\000\055\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\055\000\
    \061\000\000\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\000\000\055\000\006\000\055\000\
    \000\000\006\000\006\000\006\000\000\000\000\000\000\000\006\000\
    \006\000\000\000\006\000\054\000\006\000\201\000\201\000\201\000\
    \201\000\201\000\201\000\201\000\201\000\201\000\201\000\006\000\
    \000\000\006\000\006\000\006\000\006\000\006\000\201\000\201\000\
    \201\000\201\000\201\000\201\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\053\000\006\000\000\000\000\000\201\000\201\000\
    \201\000\201\000\201\000\201\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\006\000\052\000\006\000\000\000\000\000\000\000\000\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \000\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\055\000\000\000\000\000\055\000\055\000\055\000\000\000\
    \000\000\000\000\055\000\055\000\000\000\055\000\055\000\055\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\055\000\000\000\055\000\055\000\055\000\055\000\
    \055\000\000\000\215\000\000\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\055\000\057\000\
    \214\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\000\000\055\000\049\000\055\000\000\000\
    \049\000\049\000\049\000\000\000\000\000\000\000\049\000\049\000\
    \000\000\049\000\048\000\049\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\049\000\000\000\
    \051\000\049\000\049\000\049\000\049\000\000\000\165\000\000\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\049\000\164\000\165\000\000\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \049\000\164\000\050\000\000\000\000\000\000\000\000\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\000\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\049\000\000\000\
    \000\000\049\000\049\000\049\000\000\000\000\000\000\000\049\000\
    \049\000\000\000\049\000\049\000\049\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\049\000\
    \255\255\049\000\049\000\049\000\049\000\049\000\000\000\000\000\
    \000\000\006\000\000\000\000\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\006\000\006\000\000\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\006\000\049\000\006\000\006\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\055\000\000\000\000\000\055\000\
    \055\000\055\000\000\000\000\000\000\000\055\000\055\000\000\000\
    \055\000\055\000\055\000\000\000\000\000\175\000\000\000\175\000\
    \000\000\049\000\000\000\049\000\175\000\055\000\006\000\055\000\
    \055\000\055\000\055\000\055\000\000\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\174\000\174\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\006\000\000\000\006\000\000\000\
    \000\000\055\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\175\000\000\000\000\000\000\000\000\000\000\000\175\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\055\000\
    \000\000\055\000\000\000\175\000\000\000\000\000\000\000\175\000\
    \000\000\175\000\000\000\000\000\000\000\173\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\202\000\202\000\202\000\202\000\202\000\202\000\202\000\
    \202\000\202\000\202\000\059\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\202\000\202\000\202\000\202\000\202\000\202\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\202\000\202\000\202\000\202\000\202\000\202\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\204\000\204\000\204\000\204\000\204\000\204\000\204\000\
    \204\000\204\000\204\000\058\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\204\000\204\000\204\000\204\000\204\000\204\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\204\000\204\000\204\000\204\000\204\000\204\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\064\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\063\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\055\000\000\000\000\000\
    \055\000\055\000\055\000\000\000\000\000\000\000\055\000\055\000\
    \000\000\055\000\055\000\055\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\055\000\255\255\
    \055\000\055\000\055\000\055\000\055\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\215\000\055\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\000\000\000\000\214\000\
    \055\000\000\000\055\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\066\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\067\000\255\255\067\000\000\000\255\255\255\255\067\000\
    \067\000\255\255\067\000\255\255\067\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\067\000\067\000\067\000\255\255\067\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\067\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\067\000\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\000\000\000\000\000\000\069\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\070\000\
    \255\255\070\000\000\000\255\255\255\255\070\000\070\000\255\255\
    \070\000\255\255\070\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\255\255\070\000\
    \070\000\070\000\255\255\070\000\000\000\000\000\000\000\000\000\
    \000\000\070\000\000\000\000\000\070\000\070\000\070\000\000\000\
    \000\000\000\000\070\000\070\000\000\000\070\000\000\000\070\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\070\000\070\000\255\255\000\000\070\000\070\000\070\000\
    \070\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\070\000\
    \255\255\255\255\255\255\255\255\000\000\000\000\070\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\070\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\088\000\255\255\088\000\255\255\
    \255\255\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\204\000\204\000\204\000\204\000\204\000\204\000\
    \204\000\204\000\204\000\204\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\204\000\204\000\204\000\204\000\204\000\
    \204\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\204\000\204\000\204\000\204\000\204\000\
    \204\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\205\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \072\000\000\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\255\255\255\255\
    \255\255\255\255\073\000\255\255\000\000\000\000\000\000\000\000\
    \072\000\255\255\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\086\000\086\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \078\000\078\000\078\000\078\000\078\000\078\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \078\000\078\000\078\000\078\000\078\000\078\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\079\000\255\255\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \078\000\078\000\078\000\078\000\078\000\078\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\080\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\255\255\255\255\255\255\255\255\078\000\255\255\
    \078\000\078\000\078\000\078\000\078\000\078\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\080\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\079\000\079\000\079\000\079\000\079\000\079\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \084\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\255\255\255\255\255\255\255\255\079\000\
    \255\255\079\000\079\000\079\000\079\000\079\000\079\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \084\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\082\000\255\255\082\000\255\255\255\255\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\255\255\255\255\255\255\
    \255\255\081\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \255\255\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\082\000\255\255\082\000\255\255\255\255\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\255\255\255\255\255\255\255\255\085\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\086\000\086\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\255\255\255\255\255\255\
    \255\255\086\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \255\255\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\255\255\255\255\255\255\
    \255\255\087\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\102\000\116\000\102\000\000\000\116\000\
    \116\000\116\000\102\000\000\000\000\000\116\000\116\000\000\000\
    \116\000\116\000\116\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\100\000\100\000\100\000\116\000\000\000\116\000\
    \116\000\116\000\116\000\116\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\118\000\000\000\118\000\118\000\118\000\
    \118\000\000\000\000\000\000\000\118\000\118\000\000\000\118\000\
    \118\000\118\000\000\000\000\000\000\000\000\000\000\000\102\000\
    \000\000\116\000\000\000\000\000\118\000\102\000\118\000\118\000\
    \118\000\118\000\118\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\102\000\101\000\000\000\000\000\102\000\000\000\102\000\
    \000\000\006\000\000\000\099\000\006\000\006\000\006\000\116\000\
    \000\000\116\000\006\000\006\000\000\000\006\000\006\000\006\000\
    \118\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\006\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \130\000\000\000\255\255\130\000\130\000\130\000\118\000\000\000\
    \118\000\130\000\130\000\000\000\130\000\130\000\130\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\006\000\000\000\
    \000\000\130\000\000\000\130\000\130\000\130\000\130\000\130\000\
    \000\000\000\000\000\000\130\000\000\000\000\000\130\000\130\000\
    \130\000\000\000\000\000\000\000\130\000\130\000\000\000\130\000\
    \130\000\130\000\000\000\000\000\006\000\000\000\006\000\000\000\
    \000\000\000\000\000\000\000\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\130\000\000\000\000\000\130\000\130\000\130\000\000\000\
    \000\000\000\000\130\000\130\000\000\000\130\000\130\000\130\000\
    \000\000\000\000\000\000\130\000\000\000\130\000\000\000\000\000\
    \130\000\000\000\130\000\255\255\130\000\130\000\130\000\130\000\
    \130\000\000\000\000\000\000\000\006\000\000\000\000\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\006\000\006\000\000\000\
    \006\000\006\000\006\000\000\000\000\000\000\000\130\000\000\000\
    \130\000\000\000\000\000\000\000\000\000\006\000\130\000\006\000\
    \006\000\006\000\006\000\006\000\000\000\000\000\000\000\006\000\
    \000\000\000\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \006\000\006\000\000\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\130\000\000\000\130\000\000\000\
    \006\000\006\000\006\000\006\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\148\000\000\000\000\000\148\000\148\000\148\000\
    \000\000\000\000\000\000\148\000\148\000\000\000\148\000\148\000\
    \148\000\000\000\000\000\000\000\000\000\000\000\000\000\006\000\
    \000\000\006\000\000\000\148\000\006\000\148\000\148\000\148\000\
    \148\000\148\000\000\000\000\000\000\000\148\000\000\000\000\000\
    \148\000\148\000\148\000\000\000\000\000\000\000\148\000\148\000\
    \000\000\148\000\148\000\148\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\000\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\000\000\000\000\000\000\
    \130\000\000\000\000\000\130\000\130\000\130\000\000\000\000\000\
    \000\000\130\000\130\000\000\000\130\000\130\000\130\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\148\000\000\000\148\000\
    \000\000\130\000\148\000\130\000\130\000\130\000\130\000\130\000\
    \000\000\000\000\000\000\130\000\000\000\000\000\130\000\130\000\
    \130\000\000\000\000\000\000\000\130\000\130\000\000\000\130\000\
    \130\000\130\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \148\000\000\000\148\000\000\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\000\000\196\000\000\000\000\000\197\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\195\000\130\000\195\000\130\000\000\000\000\000\
    \130\000\195\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\194\000\194\000\194\000\194\000\194\000\194\000\
    \194\000\194\000\194\000\194\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\130\000\000\000\
    \130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\195\000\000\000\
    \000\000\000\000\000\000\000\000\195\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \195\000\000\000\000\000\000\000\195\000\000\000\195\000\192\000\
    \000\000\000\000\193\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\043\000\000\000\000\000\043\000\044\000\
    \046\000\047\000\044\000\046\000\047\000\095\000\122\000\128\000\
    \095\000\122\000\128\000\155\000\169\000\223\000\155\000\169\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\015\000\
    \012\000\016\000\024\000\041\000\016\000\016\000\041\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\093\000\094\000\101\000\101\000\101\000\101\000\012\000\
    \102\000\012\000\041\000\012\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\100\000\100\000\
    \100\000\100\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \105\000\105\000\105\000\105\000\105\000\105\000\105\000\105\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\106\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\109\000\112\000\003\000\003\000\003\000\114\000\
    \138\000\025\000\003\000\003\000\025\000\003\000\003\000\003\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\
    \108\000\108\000\003\000\139\000\003\000\003\000\003\000\003\000\
    \003\000\143\000\123\000\160\000\004\000\123\000\025\000\004\000\
    \004\000\004\000\161\000\167\000\124\000\004\000\004\000\124\000\
    \004\000\004\000\004\000\168\000\175\000\126\000\125\000\125\000\
    \126\000\157\000\125\000\178\000\157\000\004\000\003\000\004\000\
    \004\000\004\000\004\000\004\000\124\000\181\000\183\000\005\000\
    \192\000\183\000\005\000\005\000\005\000\125\000\216\000\125\000\
    \005\000\005\000\196\000\005\000\005\000\005\000\157\000\188\000\
    \217\000\218\000\188\000\025\000\003\000\220\000\003\000\222\000\
    \005\000\004\000\005\000\005\000\005\000\005\000\005\000\255\255\
    \197\000\196\000\006\000\197\000\219\000\006\000\006\000\006\000\
    \255\255\221\000\184\000\006\000\006\000\184\000\006\000\006\000\
    \006\000\198\000\206\000\212\000\198\000\206\000\212\000\004\000\
    \255\255\004\000\255\255\006\000\005\000\006\000\006\000\006\000\
    \006\000\006\000\184\000\157\000\207\000\007\000\255\255\207\000\
    \007\000\007\000\007\000\255\255\255\255\213\000\007\000\007\000\
    \213\000\007\000\007\000\007\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\005\000\255\255\007\000\006\000\
    \007\000\007\000\007\000\007\000\007\000\255\255\255\255\255\255\
    \008\000\255\255\255\255\008\000\008\000\008\000\255\255\255\255\
    \255\255\008\000\008\000\255\255\008\000\008\000\008\000\255\255\
    \255\255\255\255\255\255\255\255\184\000\006\000\255\255\006\000\
    \255\255\008\000\007\000\008\000\008\000\008\000\008\000\008\000\
    \255\255\255\255\255\255\010\000\255\255\255\255\010\000\010\000\
    \010\000\255\255\255\255\255\255\010\000\010\000\255\255\010\000\
    \010\000\010\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \007\000\255\255\007\000\255\255\010\000\008\000\010\000\010\000\
    \010\000\010\000\010\000\255\255\255\255\255\255\255\255\255\255\
    \013\000\255\255\255\255\013\000\013\000\013\000\207\000\025\000\
    \255\255\013\000\013\000\255\255\013\000\013\000\013\000\255\255\
    \255\255\255\255\255\255\008\000\255\255\008\000\255\255\010\000\
    \010\000\013\000\255\255\013\000\013\000\013\000\013\000\013\000\
    \123\000\255\255\255\255\014\000\255\255\255\255\014\000\014\000\
    \014\000\255\255\124\000\255\255\014\000\014\000\255\255\014\000\
    \014\000\014\000\255\255\126\000\255\255\125\000\010\000\157\000\
    \010\000\255\255\255\255\255\255\014\000\013\000\014\000\014\000\
    \014\000\014\000\014\000\255\255\255\255\255\255\017\000\255\255\
    \255\255\017\000\017\000\017\000\255\255\255\255\255\255\017\000\
    \017\000\255\255\017\000\017\000\017\000\255\255\255\255\218\000\
    \255\255\255\255\255\255\013\000\255\255\013\000\255\255\017\000\
    \014\000\017\000\017\000\017\000\017\000\017\000\255\255\255\255\
    \255\255\021\000\219\000\255\255\021\000\021\000\021\000\221\000\
    \184\000\255\255\021\000\021\000\255\255\021\000\021\000\021\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\014\000\255\255\
    \014\000\255\255\021\000\017\000\021\000\021\000\021\000\021\000\
    \021\000\255\255\207\000\255\255\255\255\255\255\255\255\255\255\
    \120\000\255\255\255\255\255\255\255\255\255\255\067\000\255\255\
    \255\255\067\000\067\000\067\000\255\255\255\255\255\255\067\000\
    \067\000\017\000\067\000\017\000\067\000\255\255\021\000\120\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\067\000\
    \022\000\255\255\067\000\067\000\067\000\067\000\255\255\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\255\255\255\255\255\255\021\000\255\255\021\000\022\000\
    \022\000\255\255\022\000\022\000\022\000\022\000\255\255\255\255\
    \255\255\022\000\022\000\067\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\255\255\022\000\022\000\022\000\022\000\022\000\
    \255\255\255\255\255\255\023\000\255\255\255\255\023\000\023\000\
    \023\000\067\000\255\255\023\000\023\000\023\000\255\255\023\000\
    \023\000\023\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\023\000\022\000\023\000\023\000\
    \023\000\023\000\023\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\194\000\194\000\194\000\
    \194\000\194\000\194\000\194\000\194\000\194\000\194\000\255\255\
    \255\255\255\255\255\255\022\000\026\000\022\000\255\255\091\000\
    \023\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\255\255\091\000\255\255\023\000\026\000\
    \023\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\255\255\026\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\255\255\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\255\255\255\255\
    \255\255\255\255\028\000\255\255\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\255\255\255\255\
    \255\255\255\255\028\000\255\255\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\255\255\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\255\255\029\000\
    \255\255\255\255\029\000\255\255\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\255\255\029\000\
    \255\255\255\255\029\000\255\255\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\028\000\255\255\255\255\
    \255\255\255\255\255\255\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\255\255\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\099\000\099\000\099\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\199\000\199\000\199\000\199\000\199\000\199\000\
    \199\000\199\000\199\000\199\000\255\255\255\255\255\255\255\255\
    \255\255\152\000\255\255\255\255\152\000\255\255\030\000\030\000\
    \030\000\030\000\255\255\030\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\152\000\255\255\255\255\255\255\255\255\152\000\152\000\
    \255\255\152\000\255\255\255\255\255\255\029\000\030\000\030\000\
    \030\000\030\000\030\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\255\255\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\111\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\111\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\152\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\031\000\031\000\
    \031\000\031\000\255\255\031\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\031\000\031\000\
    \031\000\031\000\031\000\030\000\255\255\255\255\255\255\255\255\
    \255\255\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \255\255\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\119\000\119\000\255\255\255\255\119\000\152\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\119\000\255\255\119\000\032\000\032\000\032\000\032\000\
    \255\255\032\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\031\000\032\000\032\000\032\000\032\000\
    \032\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \255\255\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\033\000\033\000\033\000\033\000\
    \255\255\033\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\033\000\033\000\033\000\033\000\
    \033\000\032\000\255\255\255\255\255\255\255\255\255\255\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\255\255\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\255\255\
    \255\255\119\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\180\000\180\000\180\000\180\000\180\000\180\000\255\255\
    \255\255\255\255\034\000\034\000\034\000\034\000\255\255\034\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\034\000\255\255\255\255\
    \255\255\180\000\180\000\180\000\180\000\180\000\180\000\255\255\
    \255\255\033\000\034\000\034\000\034\000\034\000\034\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\255\255\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\036\000\255\255\255\255\036\000\036\000\036\000\
    \255\255\255\255\255\255\036\000\036\000\255\255\036\000\036\000\
    \036\000\255\255\035\000\035\000\035\000\035\000\255\255\035\000\
    \255\255\255\255\255\255\036\000\035\000\036\000\036\000\036\000\
    \036\000\036\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\035\000\035\000\035\000\035\000\035\000\034\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\036\000\
    \036\000\255\255\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\255\255\036\000\039\000\036\000\
    \255\255\039\000\039\000\039\000\255\255\255\255\255\255\039\000\
    \039\000\255\255\039\000\039\000\039\000\193\000\193\000\193\000\
    \193\000\193\000\193\000\193\000\193\000\193\000\193\000\039\000\
    \255\255\039\000\039\000\039\000\039\000\039\000\193\000\193\000\
    \193\000\193\000\193\000\193\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\039\000\039\000\255\255\255\255\193\000\193\000\
    \193\000\193\000\193\000\193\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\035\000\
    \255\255\039\000\039\000\039\000\255\255\255\255\255\255\255\255\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \255\255\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\037\000\255\255\255\255\037\000\037\000\037\000\255\255\
    \255\255\255\255\037\000\037\000\255\255\037\000\037\000\037\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\037\000\255\255\037\000\037\000\037\000\037\000\
    \037\000\255\255\209\000\255\255\209\000\209\000\209\000\209\000\
    \209\000\209\000\209\000\209\000\209\000\209\000\209\000\209\000\
    \209\000\209\000\209\000\209\000\209\000\209\000\209\000\209\000\
    \209\000\209\000\209\000\209\000\209\000\209\000\037\000\037\000\
    \209\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\255\255\037\000\040\000\037\000\255\255\
    \040\000\040\000\040\000\255\255\255\255\255\255\040\000\040\000\
    \255\255\040\000\040\000\040\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\040\000\255\255\
    \040\000\040\000\040\000\040\000\040\000\255\255\158\000\255\255\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\040\000\158\000\165\000\255\255\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \040\000\165\000\040\000\255\255\255\255\255\255\255\255\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\255\255\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\255\255\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\255\255\255\255\255\255\255\255\255\255\049\000\255\255\
    \255\255\049\000\049\000\049\000\255\255\255\255\255\255\049\000\
    \049\000\255\255\049\000\049\000\049\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\038\000\038\000\038\000\038\000\049\000\
    \038\000\049\000\049\000\049\000\049\000\049\000\255\255\255\255\
    \255\255\054\000\255\255\255\255\054\000\054\000\054\000\255\255\
    \255\255\255\255\054\000\054\000\255\255\054\000\054\000\054\000\
    \255\255\255\255\255\255\038\000\038\000\038\000\038\000\038\000\
    \255\255\255\255\054\000\049\000\054\000\054\000\054\000\054\000\
    \054\000\255\255\255\255\255\255\055\000\255\255\255\255\055\000\
    \055\000\055\000\255\255\255\255\255\255\055\000\055\000\255\255\
    \055\000\055\000\055\000\255\255\255\255\166\000\255\255\166\000\
    \255\255\049\000\255\255\049\000\166\000\055\000\054\000\055\000\
    \055\000\055\000\055\000\055\000\255\255\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\166\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\054\000\255\255\054\000\255\255\
    \255\255\055\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\166\000\255\255\255\255\255\255\255\255\255\255\166\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\055\000\
    \255\255\055\000\255\255\166\000\255\255\255\255\255\255\166\000\
    \255\255\166\000\255\255\255\255\255\255\166\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \038\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \255\255\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\201\000\201\000\201\000\201\000\201\000\201\000\201\000\
    \201\000\201\000\201\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\201\000\201\000\201\000\201\000\201\000\201\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\056\000\056\000\056\000\056\000\
    \255\255\056\000\201\000\201\000\201\000\201\000\201\000\201\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\056\000\056\000\056\000\056\000\
    \056\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \255\255\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\203\000\203\000\203\000\203\000\203\000\203\000\203\000\
    \203\000\203\000\203\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\203\000\203\000\203\000\203\000\203\000\203\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\057\000\057\000\057\000\057\000\
    \255\255\057\000\203\000\203\000\203\000\203\000\203\000\203\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\057\000\057\000\057\000\057\000\
    \057\000\056\000\255\255\255\255\255\255\255\255\255\255\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\255\255\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\060\000\060\000\060\000\060\000\255\255\060\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\057\000\060\000\060\000\060\000\060\000\060\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\255\255\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \255\255\255\255\255\255\255\255\255\255\062\000\255\255\255\255\
    \062\000\062\000\062\000\255\255\255\255\255\255\062\000\062\000\
    \255\255\062\000\062\000\062\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\061\000\061\000\061\000\061\000\062\000\061\000\
    \062\000\062\000\062\000\062\000\062\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\061\000\061\000\061\000\061\000\061\000\060\000\
    \255\255\215\000\062\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\255\255\255\255\215\000\
    \062\000\255\255\062\000\255\255\255\255\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\255\255\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\061\000\
    \065\000\065\000\065\000\065\000\255\255\065\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\065\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \065\000\065\000\065\000\065\000\065\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\255\255\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \066\000\066\000\066\000\066\000\255\255\066\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \066\000\066\000\066\000\066\000\066\000\065\000\255\255\255\255\
    \255\255\255\255\255\255\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\255\255\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\068\000\068\000\
    \068\000\068\000\255\255\068\000\255\255\255\255\255\255\068\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\066\000\068\000\068\000\
    \068\000\068\000\068\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\255\255\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\255\255\255\255\255\255\255\255\
    \255\255\070\000\255\255\255\255\070\000\070\000\070\000\255\255\
    \255\255\255\255\070\000\070\000\255\255\070\000\255\255\070\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\069\000\069\000\
    \069\000\069\000\070\000\069\000\255\255\070\000\070\000\070\000\
    \070\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\069\000\069\000\
    \069\000\069\000\069\000\068\000\255\255\255\255\070\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\070\000\255\255\255\255\255\255\
    \255\255\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \255\255\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\069\000\071\000\071\000\071\000\071\000\
    \255\255\071\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\071\000\071\000\071\000\071\000\
    \071\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \255\255\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\204\000\204\000\204\000\204\000\204\000\204\000\
    \204\000\204\000\204\000\204\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\204\000\204\000\204\000\204\000\204\000\
    \204\000\255\255\255\255\255\255\072\000\072\000\072\000\072\000\
    \255\255\072\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\204\000\204\000\204\000\204\000\204\000\
    \204\000\255\255\255\255\255\255\072\000\072\000\072\000\072\000\
    \072\000\071\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\204\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\255\255\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\255\255\255\255\255\255\255\255\
    \073\000\255\255\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\255\255\255\255\255\255\255\255\
    \073\000\072\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\255\255\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\074\000\074\000\
    \074\000\074\000\255\255\074\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\074\000\074\000\
    \074\000\074\000\074\000\073\000\255\255\255\255\255\255\255\255\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\075\000\075\000\075\000\075\000\
    \255\255\075\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\074\000\075\000\075\000\075\000\075\000\
    \075\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \255\255\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\255\255\255\255\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\076\000\076\000\076\000\076\000\
    \255\255\076\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\076\000\076\000\076\000\076\000\
    \076\000\075\000\255\255\255\255\255\255\255\255\255\255\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\255\255\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\077\000\077\000\077\000\077\000\255\255\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\076\000\077\000\077\000\077\000\077\000\077\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\255\255\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\077\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\078\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\255\255\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\080\000\080\000\080\000\080\000\255\255\
    \080\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\080\000\080\000\080\000\080\000\080\000\
    \079\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\255\255\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\255\255\255\255\255\255\255\255\255\255\
    \080\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\255\255\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\083\000\083\000\083\000\
    \083\000\255\255\083\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\083\000\083\000\083\000\
    \083\000\083\000\081\000\255\255\255\255\255\255\255\255\255\255\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\255\255\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\084\000\084\000\084\000\084\000\255\255\
    \084\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\083\000\084\000\084\000\084\000\084\000\084\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\255\255\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \255\255\255\255\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\255\255\255\255\255\255\255\255\255\255\255\255\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\255\255\255\255\255\255\255\255\255\255\255\255\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \084\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\255\255\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\255\255\255\255\255\255\255\255\255\255\
    \085\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\255\255\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\086\000\092\000\116\000\092\000\255\255\116\000\
    \116\000\116\000\092\000\255\255\255\255\116\000\116\000\255\255\
    \116\000\116\000\116\000\092\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\116\000\255\255\116\000\
    \116\000\116\000\116\000\116\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\118\000\255\255\118\000\118\000\118\000\
    \118\000\255\255\255\255\255\255\118\000\118\000\255\255\118\000\
    \118\000\118\000\255\255\255\255\255\255\255\255\255\255\092\000\
    \255\255\116\000\255\255\255\255\118\000\092\000\118\000\118\000\
    \118\000\118\000\118\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\092\000\092\000\255\255\255\255\092\000\255\255\092\000\
    \255\255\129\000\255\255\092\000\129\000\129\000\129\000\116\000\
    \255\255\116\000\129\000\129\000\255\255\129\000\129\000\129\000\
    \118\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\129\000\255\255\129\000\129\000\129\000\129\000\
    \129\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \130\000\255\255\087\000\130\000\130\000\130\000\118\000\255\255\
    \118\000\130\000\130\000\255\255\130\000\130\000\130\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\129\000\255\255\
    \255\255\130\000\255\255\130\000\130\000\130\000\130\000\130\000\
    \255\255\255\255\255\255\131\000\255\255\255\255\131\000\131\000\
    \131\000\255\255\255\255\255\255\131\000\131\000\255\255\131\000\
    \131\000\131\000\255\255\255\255\129\000\255\255\129\000\255\255\
    \255\255\255\255\255\255\255\255\131\000\130\000\131\000\131\000\
    \131\000\131\000\131\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\132\000\255\255\255\255\132\000\132\000\132\000\255\255\
    \255\255\255\255\132\000\132\000\255\255\132\000\132\000\132\000\
    \255\255\255\255\255\255\130\000\255\255\130\000\255\255\255\255\
    \131\000\255\255\132\000\092\000\132\000\132\000\132\000\132\000\
    \132\000\255\255\255\255\255\255\137\000\255\255\255\255\137\000\
    \137\000\137\000\255\255\255\255\255\255\137\000\137\000\255\255\
    \137\000\137\000\137\000\255\255\255\255\255\255\131\000\255\255\
    \131\000\255\255\255\255\255\255\255\255\137\000\132\000\137\000\
    \137\000\137\000\137\000\137\000\255\255\255\255\255\255\147\000\
    \255\255\255\255\147\000\147\000\147\000\255\255\255\255\255\255\
    \147\000\147\000\255\255\147\000\147\000\147\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\132\000\255\255\132\000\255\255\
    \147\000\137\000\147\000\147\000\147\000\147\000\147\000\255\255\
    \255\255\255\255\148\000\255\255\255\255\148\000\148\000\148\000\
    \255\255\255\255\255\255\148\000\148\000\255\255\148\000\148\000\
    \148\000\255\255\255\255\255\255\255\255\255\255\255\255\137\000\
    \255\255\137\000\255\255\148\000\147\000\148\000\148\000\148\000\
    \148\000\148\000\255\255\255\255\255\255\149\000\255\255\255\255\
    \149\000\149\000\149\000\255\255\255\255\255\255\149\000\149\000\
    \255\255\149\000\149\000\149\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\147\000\255\255\147\000\255\255\149\000\148\000\
    \149\000\149\000\149\000\149\000\149\000\255\255\255\255\255\255\
    \150\000\255\255\255\255\150\000\150\000\150\000\255\255\255\255\
    \255\255\150\000\150\000\255\255\150\000\150\000\150\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\148\000\255\255\148\000\
    \255\255\150\000\149\000\150\000\150\000\150\000\150\000\150\000\
    \255\255\255\255\255\255\151\000\255\255\255\255\151\000\151\000\
    \151\000\255\255\255\255\255\255\151\000\151\000\255\255\151\000\
    \151\000\151\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \149\000\255\255\149\000\255\255\151\000\150\000\151\000\151\000\
    \151\000\151\000\151\000\255\255\189\000\255\255\255\255\189\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\189\000\150\000\189\000\150\000\255\255\255\255\
    \151\000\189\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\189\000\189\000\189\000\189\000\189\000\189\000\
    \189\000\189\000\189\000\189\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\151\000\255\255\
    \151\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\189\000\255\255\
    \255\255\255\255\255\255\255\255\189\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \189\000\255\255\255\255\255\255\189\000\255\255\189\000\189\000\
    \255\255\255\255\189\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\189\000";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\010\000\
    \036\000\012\000\000\000\000\000\000\000\002\000\000\000\027\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\001\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\002\000\004\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\039\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\019\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\036\000\036\000\000\000\036\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\000\000\000\000\001\000\022\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\007\000\001\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\001\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\022\000\124\000\189\000\197\000\124\000\198\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \022\000\255\255\124\000\000\000\125\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\119\000\120\000\255\255\255\255\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\120\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \124\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_code =
   "\255\004\255\255\005\255\255\007\255\006\255\255\003\255\000\004\
    \001\005\255\007\255\255\006\255\007\255\255\000\004\001\005\003\
    \006\002\007\255\001\255\255\000\001\255";
}

let rec token state lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 8 (-1); __ocaml_lex_token_rec state lexbuf 0
and __ocaml_lex_token_rec state lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 352 "src/ocaml/preprocess/408/lexer_raw.mll"
             bs
# 2220 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 352 "src/ocaml/preprocess/408/lexer_raw.mll"
                         (
      match state.preprocessor with
      | None -> fail lexbuf (Illegal_character bs)
      | Some _ ->
        update_loc lexbuf None 1 false 0;
        token state lexbuf )
# 2229 "src/ocaml/preprocess/lexer_raw.ml"

  | 1 ->
# 359 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 1 false 0;
        match state.preprocessor with
        | None -> token state lexbuf
        | Some _ -> return EOL
      )
# 2238 "src/ocaml/preprocess/lexer_raw.ml"

  | 2 ->
# 365 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( token state lexbuf )
# 2243 "src/ocaml/preprocess/lexer_raw.ml"

  | 3 ->
# 367 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return DOTLESS )
# 2248 "src/ocaml/preprocess/lexer_raw.ml"

  | 4 ->
# 369 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return (keyword_or state (Lexing.lexeme lexbuf) (INFIXOP0 ">.")) )
# 2253 "src/ocaml/preprocess/lexer_raw.ml"

  | 5 ->
# 371 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return (keyword_or state (Lexing.lexeme lexbuf) DOTTILDE) )
# 2258 "src/ocaml/preprocess/lexer_raw.ml"

  | 6 ->
# 373 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return UNDERSCORE )
# 2263 "src/ocaml/preprocess/lexer_raw.ml"

  | 7 ->
# 375 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return TILDE )
# 2268 "src/ocaml/preprocess/lexer_raw.ml"

  | 8 ->
let
# 381 "src/ocaml/preprocess/408/lexer_raw.mll"
                                  name
# 2274 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 382 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( lABEL (check_label_name lexbuf name) )
# 2278 "src/ocaml/preprocess/lexer_raw.ml"

  | 9 ->
let
# 383 "src/ocaml/preprocess/408/lexer_raw.mll"
                                                name
# 2284 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 384 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( warn_latin1 lexbuf;
        return (LABEL name) )
# 2289 "src/ocaml/preprocess/lexer_raw.ml"

  | 10 ->
# 387 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return QUESTION )
# 2294 "src/ocaml/preprocess/lexer_raw.ml"

  | 11 ->
# 389 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return QUESTIONQUESTION )
# 2299 "src/ocaml/preprocess/lexer_raw.ml"

  | 12 ->
let
# 390 "src/ocaml/preprocess/408/lexer_raw.mll"
                                  name
# 2305 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 391 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( oPTLABEL (check_label_name lexbuf name) )
# 2309 "src/ocaml/preprocess/lexer_raw.ml"

  | 13 ->
let
# 392 "src/ocaml/preprocess/408/lexer_raw.mll"
                                                name
# 2315 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 393 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( warn_latin1 lexbuf; return (OPTLABEL name) )
# 2319 "src/ocaml/preprocess/lexer_raw.ml"

  | 14 ->
let
# 394 "src/ocaml/preprocess/408/lexer_raw.mll"
                             name
# 2325 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 395 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (try Hashtbl.find state.keywords name
              with Not_found ->
              try Hashtbl.find keyword_table name
              with Not_found ->
                LIDENT name) )
# 2333 "src/ocaml/preprocess/lexer_raw.ml"

  | 15 ->
let
# 400 "src/ocaml/preprocess/408/lexer_raw.mll"
                                           name
# 2339 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 401 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( warn_latin1 lexbuf; return (LIDENT name) )
# 2343 "src/ocaml/preprocess/lexer_raw.ml"

  | 16 ->
let
# 402 "src/ocaml/preprocess/408/lexer_raw.mll"
                             name
# 2349 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 403 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( (* Capitalized keywords for OUnit *)
      return (try Hashtbl.find state.keywords name
              with Not_found ->
              try Hashtbl.find keyword_table name
              with Not_found ->
                UIDENT name) )
# 2358 "src/ocaml/preprocess/lexer_raw.ml"

  | 17 ->
let
# 409 "src/ocaml/preprocess/408/lexer_raw.mll"
                                           name
# 2364 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 410 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( warn_latin1 lexbuf; return (UIDENT name) )
# 2368 "src/ocaml/preprocess/lexer_raw.ml"

  | 18 ->
let
# 411 "src/ocaml/preprocess/408/lexer_raw.mll"
                   lit
# 2374 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 411 "src/ocaml/preprocess/408/lexer_raw.mll"
                       ( return (INT (lit, None)) )
# 2378 "src/ocaml/preprocess/lexer_raw.ml"

  | 19 ->
let
# 412 "src/ocaml/preprocess/408/lexer_raw.mll"
                    lit
# 2384 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1)
and
# 412 "src/ocaml/preprocess/408/lexer_raw.mll"
                                              modif
# 2389 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_curr_pos + -1) in
# 413 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (INT (lit, Some modif)) )
# 2393 "src/ocaml/preprocess/lexer_raw.ml"

  | 20 ->
let
# 414 "src/ocaml/preprocess/408/lexer_raw.mll"
                                         lit
# 2399 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 415 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (FLOAT (lit, None)) )
# 2403 "src/ocaml/preprocess/lexer_raw.ml"

  | 21 ->
let
# 416 "src/ocaml/preprocess/408/lexer_raw.mll"
                                          lit
# 2409 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1)
and
# 416 "src/ocaml/preprocess/408/lexer_raw.mll"
                                                                    modif
# 2414 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_curr_pos + -1) in
# 417 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (FLOAT (lit, Some modif)) )
# 2418 "src/ocaml/preprocess/lexer_raw.ml"

  | 22 ->
let
# 418 "src/ocaml/preprocess/408/lexer_raw.mll"
                                                                    invalid
# 2424 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 419 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( fail lexbuf (Invalid_literal invalid) )
# 2428 "src/ocaml/preprocess/lexer_raw.ml"

  | 23 ->
# 421 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.reset state.buffer;
        state.string_start_loc <- Location.curr lexbuf;
        string state lexbuf >>= fun () ->
        lexbuf.lex_start_p <- state.string_start_loc.Location.loc_start;
        state.string_start_loc <- Location.none;
        return (STRING (Buffer.contents state.buffer, None))
      )
# 2439 "src/ocaml/preprocess/lexer_raw.ml"

  | 24 ->
let
# 428 "src/ocaml/preprocess/408/lexer_raw.mll"
                       delim
# 2445 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 429 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.reset state.buffer;
        state.string_start_loc <- Location.curr lexbuf;
        quoted_string state delim lexbuf >>= fun () ->
        lexbuf.lex_start_p <- state.string_start_loc.Location.loc_start;
        state.string_start_loc <- Location.none;
        return (STRING (Buffer.contents state.buffer, Some delim)) )
# 2454 "src/ocaml/preprocess/lexer_raw.ml"

  | 25 ->
# 436 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( update_loc lexbuf None 1 false 1;
      (* newline is ('\013'* '\010') *)
      return (CHAR '\n') )
# 2461 "src/ocaml/preprocess/lexer_raw.ml"

  | 26 ->
let
# 439 "src/ocaml/preprocess/408/lexer_raw.mll"
                                         c
# 2467 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
# 440 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (CHAR c) )
# 2471 "src/ocaml/preprocess/lexer_raw.ml"

  | 27 ->
let
# 441 "src/ocaml/preprocess/408/lexer_raw.mll"
                                                    c
# 2477 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 442 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (CHAR (char_for_backslash c)) )
# 2481 "src/ocaml/preprocess/lexer_raw.ml"

  | 28 ->
# 444 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( char_for_octal_code state lexbuf 3 >>= fun c -> return (CHAR c) )
# 2486 "src/ocaml/preprocess/lexer_raw.ml"

  | 29 ->
# 446 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( char_for_decimal_code state lexbuf 2 >>= fun c -> return (CHAR c) )
# 2491 "src/ocaml/preprocess/lexer_raw.ml"

  | 30 ->
# 448 "src/ocaml/preprocess/408/lexer_raw.mll"
    ( return (CHAR (char_for_hexadecimal_code lexbuf 3)) )
# 2496 "src/ocaml/preprocess/lexer_raw.ml"

  | 31 ->
let
# 449 "src/ocaml/preprocess/408/lexer_raw.mll"
                    esc
# 2502 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_start_pos + 3) in
# 450 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( fail lexbuf (Illegal_escape (esc, None)) )
# 2506 "src/ocaml/preprocess/lexer_raw.ml"

  | 32 ->
# 452 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( let start_loc = Location.curr lexbuf in
        state.comment_start_loc <- [start_loc];
        Buffer.reset state.buffer;
        comment state lexbuf >>= fun end_loc ->
        let s = Buffer.contents state.buffer in
        Buffer.reset state.buffer;
        return (COMMENT (s, { start_loc with
                              Location.loc_end = end_loc.Location.loc_end }))
      )
# 2519 "src/ocaml/preprocess/lexer_raw.ml"

  | 33 ->
# 462 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( let loc = Location.curr lexbuf in
        Location.prerr_warning loc Warnings.Comment_start;
        state.comment_start_loc <- [loc];
        Buffer.reset state.buffer;
        comment state lexbuf >>= fun end_loc ->
        let s = Buffer.contents state.buffer in
        Buffer.reset state.buffer;
        return (COMMENT (s, { loc with Location.loc_end = end_loc.Location.loc_end }))
      )
# 2532 "src/ocaml/preprocess/lexer_raw.ml"

  | 34 ->
# 472 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( let loc = Location.curr lexbuf in
        Location.prerr_warning loc Warnings.Comment_not_end;
        lexbuf.Lexing.lex_curr_pos <- lexbuf.Lexing.lex_curr_pos - 1;
        let curpos = lexbuf.lex_curr_p in
        lexbuf.lex_curr_p <- { curpos with pos_cnum = curpos.pos_cnum - 1 };
        return STAR
      )
# 2543 "src/ocaml/preprocess/lexer_raw.ml"

  | 35 ->
let
# 479 "src/ocaml/preprocess/408/lexer_raw.mll"
                                   num
# 2549 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 480 "src/ocaml/preprocess/408/lexer_raw.mll"
                                            name
# 2554 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_mem.(2) in
# 482 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf name (int_of_string num) true 0;
        token state lexbuf
      )
# 2560 "src/ocaml/preprocess/lexer_raw.ml"

  | 36 ->
# 485 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return HASH )
# 2565 "src/ocaml/preprocess/lexer_raw.ml"

  | 37 ->
# 486 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return AMPERSAND )
# 2570 "src/ocaml/preprocess/lexer_raw.ml"

  | 38 ->
# 487 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return AMPERAMPER )
# 2575 "src/ocaml/preprocess/lexer_raw.ml"

  | 39 ->
# 488 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return BACKQUOTE )
# 2580 "src/ocaml/preprocess/lexer_raw.ml"

  | 40 ->
# 489 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return QUOTE )
# 2585 "src/ocaml/preprocess/lexer_raw.ml"

  | 41 ->
# 490 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LPAREN )
# 2590 "src/ocaml/preprocess/lexer_raw.ml"

  | 42 ->
# 491 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return RPAREN )
# 2595 "src/ocaml/preprocess/lexer_raw.ml"

  | 43 ->
# 492 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return STAR )
# 2600 "src/ocaml/preprocess/lexer_raw.ml"

  | 44 ->
# 493 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return COMMA )
# 2605 "src/ocaml/preprocess/lexer_raw.ml"

  | 45 ->
# 494 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return MINUSGREATER )
# 2610 "src/ocaml/preprocess/lexer_raw.ml"

  | 46 ->
# 495 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return DOT )
# 2615 "src/ocaml/preprocess/lexer_raw.ml"

  | 47 ->
let
# 496 "src/ocaml/preprocess/408/lexer_raw.mll"
                                      op
# 2621 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) lexbuf.Lexing.lex_curr_pos in
# 496 "src/ocaml/preprocess/408/lexer_raw.mll"
                                          ( return (DOTOP op) )
# 2625 "src/ocaml/preprocess/lexer_raw.ml"

  | 48 ->
# 497 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return DOTDOT )
# 2630 "src/ocaml/preprocess/lexer_raw.ml"

  | 49 ->
# 498 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return COLON )
# 2635 "src/ocaml/preprocess/lexer_raw.ml"

  | 50 ->
# 499 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return COLONCOLON )
# 2640 "src/ocaml/preprocess/lexer_raw.ml"

  | 51 ->
# 500 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return COLONEQUAL )
# 2645 "src/ocaml/preprocess/lexer_raw.ml"

  | 52 ->
# 501 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return COLONGREATER )
# 2650 "src/ocaml/preprocess/lexer_raw.ml"

  | 53 ->
# 502 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return SEMI )
# 2655 "src/ocaml/preprocess/lexer_raw.ml"

  | 54 ->
# 503 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return SEMISEMI )
# 2660 "src/ocaml/preprocess/lexer_raw.ml"

  | 55 ->
# 504 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LESS )
# 2665 "src/ocaml/preprocess/lexer_raw.ml"

  | 56 ->
# 505 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LESSMINUS )
# 2670 "src/ocaml/preprocess/lexer_raw.ml"

  | 57 ->
# 506 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return EQUAL )
# 2675 "src/ocaml/preprocess/lexer_raw.ml"

  | 58 ->
# 507 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACKET )
# 2680 "src/ocaml/preprocess/lexer_raw.ml"

  | 59 ->
# 508 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACKETBAR )
# 2685 "src/ocaml/preprocess/lexer_raw.ml"

  | 60 ->
# 509 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACKETLESS )
# 2690 "src/ocaml/preprocess/lexer_raw.ml"

  | 61 ->
# 510 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACKETGREATER )
# 2695 "src/ocaml/preprocess/lexer_raw.ml"

  | 62 ->
# 511 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return RBRACKET )
# 2700 "src/ocaml/preprocess/lexer_raw.ml"

  | 63 ->
# 512 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACE )
# 2705 "src/ocaml/preprocess/lexer_raw.ml"

  | 64 ->
# 513 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACELESS )
# 2710 "src/ocaml/preprocess/lexer_raw.ml"

  | 65 ->
# 514 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return BAR )
# 2715 "src/ocaml/preprocess/lexer_raw.ml"

  | 66 ->
# 515 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return BARBAR )
# 2720 "src/ocaml/preprocess/lexer_raw.ml"

  | 67 ->
# 516 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return BARRBRACKET )
# 2725 "src/ocaml/preprocess/lexer_raw.ml"

  | 68 ->
# 517 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return GREATER )
# 2730 "src/ocaml/preprocess/lexer_raw.ml"

  | 69 ->
# 518 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return GREATERRBRACKET )
# 2735 "src/ocaml/preprocess/lexer_raw.ml"

  | 70 ->
# 519 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return RBRACE )
# 2740 "src/ocaml/preprocess/lexer_raw.ml"

  | 71 ->
# 520 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return GREATERRBRACE )
# 2745 "src/ocaml/preprocess/lexer_raw.ml"

  | 72 ->
# 521 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACKETAT )
# 2750 "src/ocaml/preprocess/lexer_raw.ml"

  | 73 ->
# 522 "src/ocaml/preprocess/408/lexer_raw.mll"
           ( return LBRACKETATAT )
# 2755 "src/ocaml/preprocess/lexer_raw.ml"

  | 74 ->
# 523 "src/ocaml/preprocess/408/lexer_raw.mll"
           ( return LBRACKETATATAT )
# 2760 "src/ocaml/preprocess/lexer_raw.ml"

  | 75 ->
# 524 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return LBRACKETPERCENT )
# 2765 "src/ocaml/preprocess/lexer_raw.ml"

  | 76 ->
# 525 "src/ocaml/preprocess/408/lexer_raw.mll"
          ( return LBRACKETPERCENTPERCENT )
# 2770 "src/ocaml/preprocess/lexer_raw.ml"

  | 77 ->
# 526 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return BANG )
# 2775 "src/ocaml/preprocess/lexer_raw.ml"

  | 78 ->
# 527 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return (INFIXOP0 "!=") )
# 2780 "src/ocaml/preprocess/lexer_raw.ml"

  | 79 ->
# 528 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return PLUS )
# 2785 "src/ocaml/preprocess/lexer_raw.ml"

  | 80 ->
# 529 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return PLUSDOT )
# 2790 "src/ocaml/preprocess/lexer_raw.ml"

  | 81 ->
# 530 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return PLUSEQ )
# 2795 "src/ocaml/preprocess/lexer_raw.ml"

  | 82 ->
# 531 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return MINUS )
# 2800 "src/ocaml/preprocess/lexer_raw.ml"

  | 83 ->
# 532 "src/ocaml/preprocess/408/lexer_raw.mll"
         ( return MINUSDOT )
# 2805 "src/ocaml/preprocess/lexer_raw.ml"

  | 84 ->
let
# 534 "src/ocaml/preprocess/408/lexer_raw.mll"
                        op
# 2811 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 535 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (PREFIXOP op) )
# 2815 "src/ocaml/preprocess/lexer_raw.ml"

  | 85 ->
let
# 536 "src/ocaml/preprocess/408/lexer_raw.mll"
                              op
# 2821 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 537 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (PREFIXOP op) )
# 2825 "src/ocaml/preprocess/lexer_raw.ml"

  | 86 ->
let
# 538 "src/ocaml/preprocess/408/lexer_raw.mll"
                                              op
# 2831 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 539 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (keyword_or state op
                       (INFIXOP0 op)) )
# 2836 "src/ocaml/preprocess/lexer_raw.ml"

  | 87 ->
let
# 541 "src/ocaml/preprocess/408/lexer_raw.mll"
                              op
# 2842 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 542 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (INFIXOP1 op) )
# 2846 "src/ocaml/preprocess/lexer_raw.ml"

  | 88 ->
let
# 543 "src/ocaml/preprocess/408/lexer_raw.mll"
                              op
# 2852 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 544 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (INFIXOP2 op) )
# 2856 "src/ocaml/preprocess/lexer_raw.ml"

  | 89 ->
let
# 545 "src/ocaml/preprocess/408/lexer_raw.mll"
                         op
# 2862 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 546 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (INFIXOP4 op) )
# 2866 "src/ocaml/preprocess/lexer_raw.ml"

  | 90 ->
# 547 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return PERCENT )
# 2871 "src/ocaml/preprocess/lexer_raw.ml"

  | 91 ->
let
# 548 "src/ocaml/preprocess/408/lexer_raw.mll"
                                  op
# 2877 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 549 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (INFIXOP3 op) )
# 2881 "src/ocaml/preprocess/lexer_raw.ml"

  | 92 ->
let
# 550 "src/ocaml/preprocess/408/lexer_raw.mll"
                                op
# 2887 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 551 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (try Hashtbl.find state.keywords op
                      with Not_found -> HASHOP op) )
# 2892 "src/ocaml/preprocess/lexer_raw.ml"

  | 93 ->
let
# 554 "src/ocaml/preprocess/408/lexer_raw.mll"
                                op
# 2898 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 555 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (try Hashtbl.find state.keywords op
                      with Not_found -> HASHOP op) )
# 2903 "src/ocaml/preprocess/lexer_raw.ml"

  | 94 ->
let
# 557 "src/ocaml/preprocess/408/lexer_raw.mll"
                                       op
# 2909 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 558 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (LETOP op) )
# 2913 "src/ocaml/preprocess/lexer_raw.ml"

  | 95 ->
let
# 559 "src/ocaml/preprocess/408/lexer_raw.mll"
                                       op
# 2919 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 560 "src/ocaml/preprocess/408/lexer_raw.mll"
            ( return (ANDOP op) )
# 2923 "src/ocaml/preprocess/lexer_raw.ml"

  | 96 ->
# 561 "src/ocaml/preprocess/408/lexer_raw.mll"
        ( return EOF )
# 2928 "src/ocaml/preprocess/lexer_raw.ml"

  | 97 ->
let
# 563 "src/ocaml/preprocess/408/lexer_raw.mll"
         illegal_char
# 2934 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 564 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( fail lexbuf (Illegal_character illegal_char) )
# 2938 "src/ocaml/preprocess/lexer_raw.ml"

  | __ocaml_lex_state -> __ocaml_lex_refill
      (fun lexbuf -> lexbuf.Lexing.refill_buff lexbuf;
         __ocaml_lex_token_rec state lexbuf __ocaml_lex_state) lexbuf

and comment state lexbuf =
   __ocaml_lex_comment_rec state lexbuf 152
and __ocaml_lex_comment_rec state lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 568 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( state.comment_start_loc <- (Location.curr lexbuf) :: state.comment_start_loc;
      Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
      comment state lexbuf
    )
# 2954 "src/ocaml/preprocess/lexer_raw.ml"

  | 1 ->
# 573 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( match state.comment_start_loc with
        | [] -> assert false
        | [_] -> state.comment_start_loc <- []; return (Location.curr lexbuf)
        | _ :: l -> state.comment_start_loc <- l;
                  Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
                  comment state lexbuf
       )
# 2965 "src/ocaml/preprocess/lexer_raw.ml"

  | 2 ->
# 581 "src/ocaml/preprocess/408/lexer_raw.mll"
      (
        state.string_start_loc <- Location.curr lexbuf;
        Buffer.add_char state.buffer '\"';
        let buffer = state.buffer in
        state.buffer <- Buffer.create 15;
        (catch (string state lexbuf) (fun e l -> match e with
             | Unterminated_string ->
               begin match state.comment_start_loc with
                 | [] -> assert false
                 | loc :: _ ->
                   let start = List.hd (List.rev state.comment_start_loc) in
                   state.comment_start_loc <- [];
                   fail_loc (Unterminated_string_in_comment (start, l)) loc
               end
             | e -> fail_loc e l
           )
        ) >>= fun () ->
      state.string_start_loc <- Location.none;
      Buffer.add_string buffer (String.escaped (Buffer.contents state.buffer));
      state.buffer <- buffer;
      Buffer.add_char state.buffer '\"';
      comment state lexbuf )
# 2991 "src/ocaml/preprocess/lexer_raw.ml"

  | 3 ->
# 604 "src/ocaml/preprocess/408/lexer_raw.mll"
      (
        let delim = Lexing.lexeme lexbuf in
        let delim = String.sub delim ~pos:1 ~len:(String.length delim - 2) in
        state.string_start_loc <- Location.curr lexbuf;
        Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
        (catch (quoted_string state delim lexbuf) (fun e l -> match e with
             | Unterminated_string ->
               begin match state.comment_start_loc with
                 | [] -> assert false
                 | loc :: _ ->
                   let start = List.hd (List.rev state.comment_start_loc) in
                   state.comment_start_loc <- [];
                   fail_loc (Unterminated_string_in_comment (start, l)) loc
               end
             | e -> fail_loc e l
           )
        ) >>= fun () ->
        state.string_start_loc <- Location.none;
        Buffer.add_char state.buffer '|';
        Buffer.add_string state.buffer delim;
        Buffer.add_char state.buffer '}';
        comment state lexbuf )
# 3017 "src/ocaml/preprocess/lexer_raw.ml"

  | 4 ->
# 628 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_string state.buffer (Lexing.lexeme lexbuf); comment state lexbuf )
# 3022 "src/ocaml/preprocess/lexer_raw.ml"

  | 5 ->
# 630 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 1 false 1;
        Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
        comment state lexbuf
      )
# 3030 "src/ocaml/preprocess/lexer_raw.ml"

  | 6 ->
# 635 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_string state.buffer (Lexing.lexeme lexbuf); comment state lexbuf )
# 3035 "src/ocaml/preprocess/lexer_raw.ml"

  | 7 ->
# 637 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_string state.buffer (Lexing.lexeme lexbuf); comment state lexbuf )
# 3040 "src/ocaml/preprocess/lexer_raw.ml"

  | 8 ->
# 639 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_string state.buffer (Lexing.lexeme lexbuf); comment state lexbuf )
# 3045 "src/ocaml/preprocess/lexer_raw.ml"

  | 9 ->
# 641 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_string state.buffer (Lexing.lexeme lexbuf); comment state lexbuf )
# 3050 "src/ocaml/preprocess/lexer_raw.ml"

  | 10 ->
# 643 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( match state.comment_start_loc with
        | [] -> assert false
        | loc :: _ ->
          let start = List.hd (List.rev state.comment_start_loc) in
          state.comment_start_loc <- [];
          fail_loc (Unterminated_comment start) loc
      )
# 3061 "src/ocaml/preprocess/lexer_raw.ml"

  | 11 ->
# 651 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 1 false 0;
        Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
        comment state lexbuf
      )
# 3069 "src/ocaml/preprocess/lexer_raw.ml"

  | 12 ->
# 656 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_string state.buffer (Lexing.lexeme lexbuf); comment state lexbuf )
# 3074 "src/ocaml/preprocess/lexer_raw.ml"

  | __ocaml_lex_state -> __ocaml_lex_refill
      (fun lexbuf -> lexbuf.Lexing.refill_buff lexbuf;
         __ocaml_lex_comment_rec state lexbuf __ocaml_lex_state) lexbuf

and string state lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1); __ocaml_lex_string_rec state lexbuf 184
and __ocaml_lex_string_rec state lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 660 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( return () )
# 3087 "src/ocaml/preprocess/lexer_raw.ml"

  | 1 ->
let
# 661 "src/ocaml/preprocess/408/lexer_raw.mll"
                                  space
# 3093 "src/ocaml/preprocess/lexer_raw.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 662 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 1 false (String.length space);
        string state lexbuf
      )
# 3099 "src/ocaml/preprocess/lexer_raw.ml"

  | 2 ->
# 666 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_char state.buffer
          (char_for_backslash (Lexing.lexeme_char lexbuf 1));
        string state lexbuf )
# 3106 "src/ocaml/preprocess/lexer_raw.ml"

  | 3 ->
# 670 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( char_for_decimal_code state lexbuf 1 >>= fun c ->
        Buffer.add_char state.buffer c;
        string state lexbuf )
# 3113 "src/ocaml/preprocess/lexer_raw.ml"

  | 4 ->
# 674 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_char state.buffer (char_for_hexadecimal_code lexbuf 2);
        string state lexbuf )
# 3119 "src/ocaml/preprocess/lexer_raw.ml"

  | 5 ->
# 677 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( store_escaped_uchar state lexbuf (uchar_for_uchar_escape lexbuf);
        string state lexbuf )
# 3125 "src/ocaml/preprocess/lexer_raw.ml"

  | 6 ->
# 680 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( if in_comment state
        then string state lexbuf
        else begin
(*  Should be an error, but we are very lax.
                  fail (Illegal_escape (Lexing.lexeme lexbuf),
                        (Location.curr lexbuf)
*)
          let loc = Location.curr lexbuf in
          Location.prerr_warning loc Warnings.Illegal_backslash;
          Buffer.add_char state.buffer (Lexing.lexeme_char lexbuf 0);
          Buffer.add_char state.buffer (Lexing.lexeme_char lexbuf 1);
          string state lexbuf
        end
      )
# 3143 "src/ocaml/preprocess/lexer_raw.ml"

  | 7 ->
# 695 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( if not (in_comment state) then
          Location.prerr_warning (Location.curr lexbuf) Warnings.Eol_in_string;
        update_loc lexbuf None 1 false 0;
        Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
        string state lexbuf
      )
# 3153 "src/ocaml/preprocess/lexer_raw.ml"

  | 8 ->
# 702 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( let loc = state.string_start_loc in
        state.string_start_loc <- Location.none;
        fail_loc Unterminated_string loc )
# 3160 "src/ocaml/preprocess/lexer_raw.ml"

  | 9 ->
# 706 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_char state.buffer (Lexing.lexeme_char lexbuf 0);
        string state lexbuf )
# 3166 "src/ocaml/preprocess/lexer_raw.ml"

  | __ocaml_lex_state -> __ocaml_lex_refill
      (fun lexbuf -> lexbuf.Lexing.refill_buff lexbuf;
         __ocaml_lex_string_rec state lexbuf __ocaml_lex_state) lexbuf

and quoted_string state delim lexbuf =
   __ocaml_lex_quoted_string_rec state delim lexbuf 207
and __ocaml_lex_quoted_string_rec state delim lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 711 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 1 false 0;
        Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
        quoted_string state delim lexbuf
      )
# 3182 "src/ocaml/preprocess/lexer_raw.ml"

  | 1 ->
# 716 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( let loc = state.string_start_loc in
        state.string_start_loc <- Location.none;
        fail_loc Unterminated_string loc )
# 3189 "src/ocaml/preprocess/lexer_raw.ml"

  | 2 ->
# 720 "src/ocaml/preprocess/408/lexer_raw.mll"
      (
        let edelim = Lexing.lexeme lexbuf in
        let edelim = String.sub edelim ~pos:1 ~len:(String.length edelim - 2) in
        if delim = edelim then return ()
        else (Buffer.add_string state.buffer (Lexing.lexeme lexbuf);
              quoted_string state delim lexbuf)
      )
# 3200 "src/ocaml/preprocess/lexer_raw.ml"

  | 3 ->
# 728 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( Buffer.add_char state.buffer (Lexing.lexeme_char lexbuf 0);
        quoted_string state delim lexbuf )
# 3206 "src/ocaml/preprocess/lexer_raw.ml"

  | __ocaml_lex_state -> __ocaml_lex_refill
      (fun lexbuf -> lexbuf.Lexing.refill_buff lexbuf;
         __ocaml_lex_quoted_string_rec state delim lexbuf __ocaml_lex_state) lexbuf

and skip_sharp_bang state lexbuf =
   __ocaml_lex_skip_sharp_bang_rec state lexbuf 216
and __ocaml_lex_skip_sharp_bang_rec state lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 733 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 3 false 0; token state lexbuf )
# 3219 "src/ocaml/preprocess/lexer_raw.ml"

  | 1 ->
# 735 "src/ocaml/preprocess/408/lexer_raw.mll"
      ( update_loc lexbuf None 1 false 0; token state lexbuf )
# 3224 "src/ocaml/preprocess/lexer_raw.ml"

  | 2 ->
# 736 "src/ocaml/preprocess/408/lexer_raw.mll"
       ( token state lexbuf )
# 3229 "src/ocaml/preprocess/lexer_raw.ml"

  | __ocaml_lex_state -> __ocaml_lex_refill
      (fun lexbuf -> lexbuf.Lexing.refill_buff lexbuf;
         __ocaml_lex_skip_sharp_bang_rec state lexbuf __ocaml_lex_state) lexbuf

;;

# 738 "src/ocaml/preprocess/408/lexer_raw.mll"
 
  type comment = string * Location.t

  (* preprocessor support not implemented, not compatible with monadic
     interface *)

  let rec token_without_comments state lexbuf =
    token state lexbuf >>= function
    | COMMENT _ ->
      token_without_comments state lexbuf
    | tok -> return tok

# 3250 "src/ocaml/preprocess/lexer_raw.ml"
